import re
from os import getenv

import click
import requests

from .base import CommentDriver
from ..terraform_out import TerraformOut


class GiteaComment(CommentDriver):
    PR_REGEX = re.compile(r"refs/pull/(\d+)/head", re.IGNORECASE)

    def add(self, workspace: str, tf_out: TerraformOut):
        gitea = Gitea(self.server, getenv('TERRATALK_GITEA_TOKEN'))
        gitea.pr(self.repository_slug, self.pull_request_id)

        for c in gitea.comments():
            if c['body'].lstrip().startswith(
                f'<!-- terratalk: {workspace} -->'
            ):
                click.echo(
                    f"[tf-comment-plan] deleting previous comment: {c['id']}"
                )
                gitea.comment_delete(c['id'])

        if not tf_out.does_nothing():
            gitea.comment_add(f'''
<!-- terratalk: {workspace} -->
### tf plan output: {workspace}
```diff
{tf_out.show()}
```
''')

    def detect(self) -> bool:
        if getenv('GITEA_ACTIONS') == 'true':
            self.type = 'gitea'
            self.server = getenv('GITHUB_API_URL')
            self.project_key = getenv('GITHUB_REPOSITORY_OWNER')
            self.repository_slug = getenv('GITHUB_REPOSITORY')
            m = self.PR_REGEX.search(getenv('GITHUB_REF', ''))
            if m:
                self.pull_request_id = int(m.group(1))
                return True
        return False


class Gitea:
    def __init__(self, server: str, token: str):
        self.server = server
        self.token = token
        self.repository_slug = None
        self.pull_request_id = None

    def pr(self, repository_slug: str, pull_request_id: int):
        self.repository_slug = repository_slug
        self.pull_request_id = pull_request_id

    def comments(self) -> list[dict]:
        return [
            c for c in self.get(
                f"repos/{self.repository_slug}/issues/comments"
            )
            if c.get('pull_request_url', '').endswith(
                f"/{self.pull_request_id}"
            )
        ]

    def comment_delete(self, comment_id: int):
        self.delete(
            f"repos/{self.repository_slug}/issues/comments/{comment_id}"
        )

    def comment_add(self, body: str):
        self.post(
            f"repos/{self.repository_slug}/issues/"
            f"{self.pull_request_id}/comments",
            {"body": body},
        )

    def get(self, path: str) -> dict:
        return requests.get(
            f"{self.server}/{path}",
            headers={"Authorization": f"token {self.token}"},
        ).json()

    def post(self, path: str, data: dict) -> dict:
        return requests.post(
            f"{self.server}/{path}",
            headers={"Authorization": f"token {self.token}"},
            json=data,
        ).json()

    def delete(self, path: str):
        return requests.delete(
            f"{self.server}/{path}",
            headers={"Authorization": f"token {self.token}"},
        )



import os
import unittest

from terratalk.terraform_out import TerraformOut


class TestTerraformOut(unittest.TestCase):

    def test_does_nothing(self):
        tf = TerraformOut('foo.plan')
        tf._raw_output = '''
No changes. Your infrastructure matches the configuration.

Terraform has compared your real infrastructure against your configuration and found no differences, so no changes are needed.
'''
        self.assertEqual(tf.does_nothing(), True)

    def test_does_nothing_0(self):
        tf = TerraformOut('foo.plan')
        tf._raw_output = '''
Plan: 0 to add, 0 to change, 0 to destroy.
'''
        self.assertEqual(tf.does_nothing(), True)

    def test_create(self):
        tf = TerraformOut('foo.plan')
        tf._raw_output = '''
Terraform used the selected providers to generate the following execution plan. Resource actions are indicated with the following symbols:
  + create

Terraform will perform the following actions:

  # aws_key_pair.deployer will be created
  + resource "aws_key_pair" "deployer" {
      + arn         = (known after apply)
      + fingerprint = (known after apply)
      + id          = (known after apply)
      + key_name    = "deployer-key"
      + key_pair_id = (known after apply)
      + public_key  = "ssh-rsa "
      + tags_all    = (known after apply)
    }

Plan: 1 to add, 0 to change, 0 to destroy.
'''

        self.assertEqual(tf.does_nothing(), False)
        self.assertEqual(tf.show(), '''
+  aws_key_pair.deployer will be created

Plan: 1 to add, 0 to change, 0 to destroy.
'''.strip())

    def test_change(self):
        tf = TerraformOut('foo.plan')
        tf._raw_output = '''
Note: Objects have changed outside of Terraform

Terraform detected the following changes made outside of Terraform since the last "terraform apply":

  # aws_key_pair.deployer has been changed
  ~ resource "aws_key_pair" "deployer" {
        id          = "deployer-key"
      + tags        = {}
        # (6 unchanged attributes hidden)
    }

Unless you have made equivalent changes to your configuration, or ignored the relevant attributes using ignore_changes, the following plan may include actions to undo or respond to these changes.

──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────

No changes. Your infrastructure matches the configuration.

Your configuration already matches the changes detected above. If you'd like to update the Terraform state to match, create and apply a refresh-only plan:
  terraform apply -refresh-only
'''

        self.assertEqual(tf.does_nothing(), False)
        self.assertEqual(tf.show(), '''
#  aws_key_pair.deployer has been changed
'''.strip())

    def test_update(self):
        tf = TerraformOut('foo.plan')
        tf._raw_output = '''

Terraform used the selected providers to generate the following execution plan. Resource actions are indicated with the following symbols:
  ~ update in-place

Terraform will perform the following actions:

  # aws_key_pair.deployer will be updated in-place
  ~ resource "aws_key_pair" "deployer" {
        id          = "deployer-key"
      ~ tags        = {
          + "Foo" = "Bar"
        }
      ~ tags_all    = {
          + "Foo" = "Bar"
        }
        # (5 unchanged attributes hidden)
    }

Plan: 0 to add, 1 to change, 0 to destroy.
'''

        self.assertEqual(tf.does_nothing(), False)
        self.assertEqual(tf.show(), '''
@  aws_key_pair.deployer will be updated in-place

Plan: 0 to add, 1 to change, 0 to destroy.
'''.strip())

    def test_destroy(self):
        tf = TerraformOut('foo.plan')
        tf._raw_output = '''
Terraform used the selected providers to generate the following execution plan. Resource actions are indicated with the following symbols:
  - destroy

Terraform will perform the following actions:

  # aws_ecs_cluster.foobar will be destroyed
  - resource "aws_ecs_cluster" "foobar" {
      - arn                = "arn:aws:ecs:eu-west-1:1234567:cluster/foobar" -> null
      - name               = "foobar" -> null
    }

Plan: 0 to add, 0 to change, 1 to destroy.

Changes to Outputs:
  - ecs_cluster_arn = "arn:aws:ecs:eu-west-1:1234567:cluster/foobar" -> null
'''

        self.assertEqual(tf.does_nothing(), False)
        self.assertEqual(tf.show(), '''
-  aws_ecs_cluster.foobar will be destroyed

Plan: 0 to add, 0 to change, 1 to destroy.
'''.strip())

    def test_replace(self):
        tf = TerraformOut('foo.plan')
        tf._raw_output = '''
  # docker_container.service must be replaced

Plan: 1 to add, 0 to change, 1 to destroy.
'''

        self.assertEqual(tf.show(), '''
-+ docker_container.service must be replaced

Plan: 1 to add, 0 to change, 1 to destroy.
'''.strip())

    def test_cli_error(self):
        tf = TerraformOut('test.plan')
        self.assertEqual(tf.show().lstrip().startswith(
            "Error: Failed to read the given file as a state or plan file"
        ), True)

    def test_cli(self):
        cwd = os.getcwd()
        os.chdir('tests')
        tf = TerraformOut('test.plan')
        self.assertEqual(tf.show(), '''
+  null_resource.foobar will be created

Plan: 1 to add, 0 to change, 0 to destroy.
'''.strip())
        os.chdir(cwd)


if __name__ == '__main__':
    unittest.main()



import re
from os import getenv

import click

from .base import CommentDriver
from ..terraform_out import TerraformOut


class GithubComment(CommentDriver):
    DETECT_REGEX = re.compile(
        r"\Ahttps://(github\.com)/([^/]+)/([^/]+)/pull/(\d+)\Z",
        re.IGNORECASE,
    )

    def add(self, workspace: str, tf_out: TerraformOut):
        from github import Github
        gh = Github(getenv('GITHUB_TOKEN'))

        repo = gh.get_repo(
            f"{self.project_key}/{self.repository_slug}"
        )
        issue = repo.get_issue(self.pull_request_id)

        # delete any older comments
        for c in issue.get_comments():
            if c.body.lstrip().startswith(f'<!-- terratalk: {workspace} -->'):
                click.echo(
                    f'[tf-comment-plan] deleting previous comment: {c.id}'
                )
                c.delete()

        if not tf_out.does_nothing():
            issue.create_comment(f'''
<!-- terratalk: {workspace} -->
### tf plan output: {workspace}
```diff
{tf_out.show()}
```
''')

    def detect(self) -> bool:
        m = self.DETECT_REGEX.search(getenv('CHANGE_URL', ''))
        if m:
            self.server = m.group(1)
            self.type = 'github'
            self.project_key = m.group(2)
            self.repository_slug = m.group(3)
            self.pull_request_id = int(m.group(4))
            return True

        return False



from abc import ABC, abstractmethod
from typing import Optional

from terratalk.terraform_out import TerraformOut


class CommentDriver(ABC):
    def __init__(self) -> None:
        self.type: Optional[str] = None
        self.server: Optional[str] = None
        self.project_key: Optional[str] = None
        self.pull_request_id: Optional[int] = None
        self.repository_slug: Optional[str] = None

    @abstractmethod
    def add(self, workspace: str, tf_out: TerraformOut):
        pass

    @abstractmethod
    def detect(self) -> bool:
        """Returns True if the environment matches this driver."""
        pass



import os
import unittest

from click.testing import CliRunner

from terratalk import __main__


class TestMain(unittest.TestCase):

    def setUp(self) -> None:
        self.cwd = os.getcwd()
        os.chdir('tests')

    def tearDown(self) -> None:
        os.chdir(self.cwd)

    def test_comment_nothing(self):
        runner = CliRunner()
        result = runner.invoke(__main__.cli, ['comment', '-w', 'nothing'])
        self.assertEqual(result.output, '[terratalk] this plan does nothing\n')

    def test_output(self):
        runner = CliRunner()
        result = runner.invoke(__main__.cli, ['output', '-w', 'test'])
        self.assertEqual(result.output, '''
+  null_resource.foobar will be created

Plan: 1 to add, 0 to change, 0 to destroy.
'''.lstrip())



import os
from unittest import TestCase, mock

from terratalk.__main__ import get_comment_driver


class TestMain(TestCase):

    @mock.patch.dict(os.environ, {
        'CHANGE_URL': 'https://bitbucket.org/test-org/test-repo/pull-requests/33',
    })
    def test_bitbucket(self):
        pr = get_comment_driver()

        self.assertEqual(pr.type, 'bitbucket')
        self.assertEqual(pr.server, 'bitbucket.org')
        self.assertEqual(pr.project_key, 'test-org')
        self.assertEqual(pr.repository_slug, 'test-repo')
        self.assertEqual(pr.pull_request_id, 33)

    @mock.patch.dict(os.environ, {
        'CHANGE_URL': 'https://stash.com/projects/FOO/repos/foobar/pull-requests/186',
    })
    def test_bitbucket_server(self):
        pr = get_comment_driver()

        self.assertEqual(pr.type, 'bitbucket-server')
        self.assertEqual(pr.server, 'https://stash.com')
        self.assertEqual(pr.project_key, 'FOO')
        self.assertEqual(pr.repository_slug, 'foobar')
        self.assertEqual(pr.pull_request_id, 186)

    @mock.patch.dict(os.environ, {
        'GITEA_ACTIONS': 'true',
        'GITHUB_API_URL': 'https://gitea.com/api/v1',
        'GITHUB_REPOSITORY_OWNER': 'foobar',
        'GITHUB_REPOSITORY': 'foobar/barfoo',
        'GITHUB_REF': 'refs/pull/1/head',
    })
    def test_gitea(self):
        pr = get_comment_driver()

        self.assertEqual(pr.type, 'gitea')
        self.assertEqual(pr.server, 'https://gitea.com/api/v1')
        self.assertEqual(pr.project_key, 'foobar')
        self.assertEqual(pr.repository_slug, 'foobar/barfoo')
        self.assertEqual(pr.pull_request_id, 1)

    @mock.patch.dict(os.environ, {
        'CHANGE_URL': 'https://github.com/test-org/test-repo/pull/6',
    })
    def test_github(self):
        pr = get_comment_driver()

        self.assertEqual(pr.type, 'github')
        self.assertEqual(pr.server, 'github.com')
        self.assertEqual(pr.project_key, 'test-org')
        self.assertEqual(pr.repository_slug, 'test-repo')
        self.assertEqual(pr.pull_request_id, 6)

    @mock.patch.dict(os.environ, {
        'CI_SERVER_NAME': 'GitLab',
        'CI_MERGE_REQUEST_IID': '8',
        'CI_MERGE_REQUEST_PROJECT_ID': '428',
        'CI_SERVER_URL': 'https://gitlab.company.com',
    })
    def test_gitlab(self):
        pr = get_comment_driver()

        self.assertEqual(pr.type, 'gitlab')
        self.assertEqual(pr.server, 'https://gitlab.company.com')
        self.assertEqual(pr.project_key, 428)
        self.assertEqual(pr.pull_request_id, 8)

    @mock.patch.dict(os.environ, {
        'CHANGE_URL': 'https://foobar.org',
    })
    def test_no_driver(self):
        self.assertRaises(RuntimeError, get_comment_driver)



import unittest

from terratalk.drivers.bitbucket_server_comment import BitbucketServer


class TestMain(unittest.TestCase):

    def test_pr(self):
        base_url = 'https://foobar.com'
        username = 'user'
        password = 'pass'
        project_key = 'FOOBAR'
        repository_slug = 'barfoo'
        pull_request_id = 11

        bs = BitbucketServer(base_url=base_url,
                             username=username,
                             password=password)

        bs.pr(
            project_key=project_key,
            repository_slug=repository_slug,
            pull_request_id=pull_request_id,
        )

        self.assertEqual(bs.base_url, f"{base_url}/rest/api/1.0")
        self.assertEqual(bs.username, username)
        self.assertEqual(bs.password, password)
        self.assertEqual(bs.project_key, project_key)
        self.assertEqual(bs.repository_slug, repository_slug)
        self.assertEqual(bs.pull_request_id, pull_request_id)



from setuptools import setup, find_packages
from distutils.util import convert_path


def find_description():
    with open(convert_path('README.md')) as fh:
        return fh.read()


setup(
    name='terratalk',
    version='0.5.1',
    py_modules=['cli'],
    author='Günter Grodotzki',
    author_email='gunter@grodotzki.com',
    description='A Terraform commentator.',
    long_description=find_description(),
    long_description_content_type='text/markdown',
    url='https://github.com/lifeofguenter/terratalk',
    packages=find_packages(exclude=['tests*']),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3',
    install_requires=[
        'Click',
        'requests',
    ],
    extras_require={
        'GitHub': ['PyGithub'],
        'GitLab': ['python-gitlab'],
    },
    entry_points={
        'console_scripts': [
            'terratalk = terratalk.__main__:cli',
        ],
    },
)



from os import getenv

import click

from .base import CommentDriver
from ..terraform_out import TerraformOut


class GitlabComment(CommentDriver):
    def add(self, workspace: str, tf_out: TerraformOut):
        import gitlab
        gl = gitlab.Gitlab(
            self.server,
            private_token=getenv('GITLAB_TOKEN'),
        )

        gl_project = gl.projects.get(self.project_key, lazy=True)
        gl_mr = gl_project.mergerequests.get(
            self.pull_request_id,
            lazy=True,
        )
        for discussion in gl_mr.discussions.list(get_all=True):
            for note in discussion.attributes['notes']:
                if note['body'].lstrip().startswith(
                    f"<!-- terratalk: {workspace} -->"
                ):
                    disc = gl_mr.discussions.get(discussion.id)
                    click.echo(
                        f"[tf-comment-plan] deleting previous comment: "
                        f"{note['id']}"
                    )
                    disc.notes.delete(id=note['id'])

        if not tf_out.does_nothing():
            gl_mr.discussions.create({'body': f'''
<!-- terratalk: {workspace} -->
### tf plan output: {workspace}
```diff
{tf_out.show()}
```
'''})

    def detect(self) -> bool:
        if getenv('CI_SERVER_NAME') == 'GitLab':
            self.type = 'gitlab'
            self.server = getenv('CI_SERVER_URL')
            self.project_key = int(getenv('CI_MERGE_REQUEST_PROJECT_ID'))
            self.pull_request_id = int(getenv('CI_MERGE_REQUEST_IID'))
            return True
        return False



import importlib
import pkgutil

import click

import terratalk.drivers
from terratalk.drivers.base import CommentDriver
from terratalk.terraform_out import TerraformOut


def get_comment_driver() -> CommentDriver:
    """Dynamically load and return the first matching PR driver."""
    for _, module_name, _ in pkgutil.iter_modules(terratalk.drivers.__path__):
        if module_name == 'base':
            continue

        module = importlib.import_module(f"terratalk.drivers.{module_name}")

        # Find a class that inherits from CommentDriver
        for obj in module.__dict__.values():
            if (isinstance(obj, type) and
               issubclass(obj, CommentDriver) and obj is not CommentDriver):
                driver_instance = obj()
                if driver_instance.detect():
                    return driver_instance

    raise RuntimeError("No PR driver matched the environment.")


@click.group()
@click.version_option()
def cli():
    pass


@cli.command()
@click.option('-w', '--workspace')
def comment(workspace):
    # fetch terraform output
    tf = TerraformOut(workspace + '.plan')

    if tf.does_nothing():
        click.echo('[terratalk] this plan does nothing')

    pr = get_comment_driver()
    pr.add(workspace, tf)


@cli.command()
@click.option('-w', '--workspace')
def output(workspace):
    tf = TerraformOut(workspace + '.plan')

    if tf.does_nothing():
        click.echo('[terratalk] this plan does nothing')
    else:
        click.echo(tf.show())


if __name__ == '__main__':
    cli()
